// $Id: icsd_radii.h,v 1.3 2001/11/26 23:05:20 rwgk Exp $
/* Copyright (c) 2001 The Regents of the University of California through
   E.O. Lawrence Berkeley National Laboratory, subject to approval by the
   U.S. Department of Energy. See files COPYRIGHT.txt and
   cctbx/LICENSE.txt for further details.

   Revision history:
     Apr 2001: SourceForge release (R.W. Grosse-Kunstleve)
 */

#ifndef CCTBX_ELTBX_ICSD_RADII_H
#define CCTBX_ELTBX_ICSD_RADII_H

#include <string>

namespace cctbx { namespace eltbx {

  namespace detail {
    struct Label_Radius {
      const char* Label;
      float       Radius;
    };
  }

  //! Access to table of ionic radii.
  /*! Reference:<pre>
                          U s e r ' s  M a n u a l
                         I C S D  -  C R Y S T I N
                         =========================
                    Inorganic Crystal Structure Database
                            in conjunction with
                   Crystal Structure Information System
                           and its application to
                CCDF - Cambridge Crystallographic Data File
                                    and
                           MDF - Metal Data File
                    G.Bergerhoff, B.Kilger, C.Witthauer,
                            R.Hundt, R.Sievers
                                    Bonn
                                    1986
             Institut fuer Anorganische Chemie der Universitaet
            -------------------------------------------------------------
            ICSD/CRYSTIN User's Manual.
            English Version. Translated by Ruth Schubert.
            Updated Dec. 1986.</pre>

      These radii are also the basis of the distance tests,  which are
      routinely  carried  out when collecting data in the ICSD system.
      In  this  connection,   negative increments are obtained for the
      specially small atoms C+4, D+1, H+1 and N+5.
   */
  class ICSD_Radius {
    public:
      //! Default constructor. Calling certain methods may cause crashes!
      ICSD_Radius() : m_Label_Radius(0)  {}
      //! Lookup ionic radius for the given ion label.
      /*! If Exact == true, the ion label must exactly
          match the tabulated label. However, the lookup is not
          case-sensitive.<br>
          E.g., "SI4" will be matched with "Si".<br>
          "Si4+" and "Si+4" will be matched with "Si4+".<br>
          See also: eltbx::StripLabel()
       */
      explicit
      ICSD_Radius(const std::string& Label, bool Exact = false);
      //! Return label from table.
      const char* Label() const { return m_Label_Radius->Label; }
      //! Return radius (Angstrom) from table.
      float Radius() const { return m_Label_Radius->Radius; }
    private:
      const detail::Label_Radius* m_Label_Radius;
  };

}} // cctbx::eltbx

#endif // CCTBX_ELTBX_ICSD_RADII_H
